import numpy as np
import pytest

import emcee
from emcee.backends.hdf import TempHDFBackend, does_hdf5_support_longdouble


def test_longdouble_doesnt_crash_bug_312():
    def log_prob(x, ivar):
        return -0.5 * np.sum(ivar * x ** 2)

    ndim, nwalkers = 5, 20
    ivar = 1.0 / np.random.rand(ndim).astype(np.longdouble)
    p0 = np.random.randn(nwalkers, ndim).astype(np.longdouble)

    sampler = emcee.EnsembleSampler(nwalkers, ndim, log_prob, args=[ivar])
    sampler.run_mcmc(p0, 100)


@pytest.mark.parametrize("cls", emcee.backends.get_test_backends())
def test_longdouble_actually_needed(cls):
    if issubclass(cls, TempHDFBackend) and not does_hdf5_support_longdouble():
        pytest.xfail("HDF5 does not support long double on this platform")

    mjd = np.longdouble(58000.0)
    sigma = 100 * np.finfo(np.longdouble).eps * mjd

    def log_prob(x):
        assert x.dtype == np.longdouble
        return -0.5 * np.sum(((x - mjd) / sigma) ** 2)

    ndim, nwalkers = 1, 20
    steps = 1000
    p0 = sigma * np.random.randn(nwalkers, ndim).astype(np.longdouble) + mjd
    assert not all(p0 == mjd)

    with cls(dtype=np.longdouble) as backend:
        sampler = emcee.EnsembleSampler(
            nwalkers, ndim, log_prob, backend=backend
        )
        sampler.run_mcmc(p0, steps)

        samples = sampler.get_chain().reshape((-1,))
        assert samples.dtype == np.longdouble

        assert not np.all(samples == mjd)
        assert np.abs(np.mean(samples) - mjd) < 10 * sigma / np.sqrt(
            len(samples)
        )
        assert 0.1 * sigma < np.std(samples) < 10 * sigma



# -*- coding: utf-8 -*-

from pkg_resources import DistributionNotFound, get_distribution

try:
    __version__ = get_distribution("emcee").version
except DistributionNotFound:
    __version__ = "unknown version"


# General stuff
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.mathjax",
    "myst_nb",
]

myst_enable_extensions = ["dollarmath", "colon_fence"]
source_suffix = ".rst"
master_doc = "index"

project = "emcee"
copyright = "2012-2021, Dan Foreman-Mackey & contributors"
version = __version__
release = __version__
exclude_patterns = ["_build"]

# HTML theme
html_theme = "sphinx_book_theme"
html_copy_source = True
html_show_sourcelink = True
html_sourcelink_suffix = ""
html_title = "emcee"
html_favicon = "_static/favicon.png"
html_static_path = ["_static"]
html_theme_options = {
    "path_to_docs": "docs",
    "repository_url": "https://github.com/dfm/emcee",
    "repository_branch": "main",
    "launch_buttons": {
        "binderhub_url": "https://mybinder.org",
        "notebook_interface": "classic",
    },
    "use_edit_page_button": True,
    "use_issues_button": True,
    "use_repository_button": True,
    "use_download_button": True,
}
jupyter_execute_notebooks = "off"
execution_timeout = -1



"""
Unit tests of some functionality in ensemble.py when the parameters are named
"""
import string
from unittest import TestCase

import numpy as np
import pytest

from emcee.ensemble import EnsembleSampler, ndarray_to_list_of_dicts


class TestNP2ListOfDicts(TestCase):
    def test_ndarray_to_list_of_dicts(self):
        # Try different numbers of keys
        for n_keys in [1, 2, 10, 26]:
            keys = list(string.ascii_lowercase[:n_keys])
            key_set = set(keys)
            key_dict = {key: i for i, key in enumerate(keys)}
            # Try different number of walker/procs
            for N in [1, 2, 3, 10, 100]:
                x = np.random.rand(N, n_keys)

                LOD = ndarray_to_list_of_dicts(x, key_dict)
                assert len(LOD) == N, "need 1 dict per row"
                for i, dct in enumerate(LOD):
                    assert dct.keys() == key_set, "keys are missing"
                    for j, key in enumerate(keys):
                        assert dct[key] == x[i, j], f"wrong value at {(i, j)}"


class TestNamedParameters(TestCase):
    """
    Test that a keyword-based log-probability function instead of
    a positional.
    """

    # Keyword based lnpdf
    def lnpdf(self, pars) -> np.float64:
        mean = pars["mean"]
        var = pars["var"]
        if var <= 0:
            return -np.inf
        return (
            -0.5 * ((mean - self.x) ** 2 / var + np.log(2 * np.pi * var)).sum()
        )

    def lnpdf_mixture(self, pars) -> np.float64:
        mean1 = pars["mean1"]
        var1 = pars["var1"]
        mean2 = pars["mean2"]
        var2 = pars["var2"]
        if var1 <= 0 or var2 <= 0:
            return -np.inf
        return (
            -0.5
            * (
                (mean1 - self.x) ** 2 / var1
                + np.log(2 * np.pi * var1)
                + (mean2 - self.x - 3) ** 2 / var2
                + np.log(2 * np.pi * var2)
            ).sum()
        )

    def lnpdf_mixture_grouped(self, pars) -> np.float64:
        mean1, mean2 = pars["means"]
        var1, var2 = pars["vars"]
        const = pars["constant"]
        if var1 <= 0 or var2 <= 0:
            return -np.inf
        return (
            -0.5
            * (
                (mean1 - self.x) ** 2 / var1
                + np.log(2 * np.pi * var1)
                + (mean2 - self.x - 3) ** 2 / var2
                + np.log(2 * np.pi * var2)
            ).sum()
            + const
        )

    def setUp(self):
        # Draw some data from a unit Gaussian
        self.x = np.random.randn(100)
        self.names = ["mean", "var"]

    def test_named_parameters(self):
        sampler = EnsembleSampler(
            nwalkers=10,
            ndim=len(self.names),
            log_prob_fn=self.lnpdf,
            parameter_names=self.names,
        )
        assert sampler.params_are_named
        assert list(sampler.parameter_names.keys()) == self.names

    def test_asserts(self):
        # ndim name mismatch
        with pytest.raises(AssertionError):
            _ = EnsembleSampler(
                nwalkers=10,
                ndim=len(self.names) - 1,
                log_prob_fn=self.lnpdf,
                parameter_names=self.names,
            )

        # duplicate names
        with pytest.raises(AssertionError):
            _ = EnsembleSampler(
                nwalkers=10,
                ndim=3,
                log_prob_fn=self.lnpdf,
                parameter_names=["a", "b", "a"],
            )

        # vectorize turned on
        with pytest.raises(AssertionError):
            _ = EnsembleSampler(
                nwalkers=10,
                ndim=len(self.names),
                log_prob_fn=self.lnpdf,
                parameter_names=self.names,
                vectorize=True,
            )

    def test_compute_log_prob(self):
        # Try different numbers of walkers
        for N in [4, 8, 10]:
            sampler = EnsembleSampler(
                nwalkers=N,
                ndim=len(self.names),
                log_prob_fn=self.lnpdf,
                parameter_names=self.names,
            )
            coords = np.random.rand(N, len(self.names))
            lnps, _ = sampler.compute_log_prob(coords)
            assert len(lnps) == N
            assert lnps.dtype == np.float64

    def test_compute_log_prob_mixture(self):
        names = ["mean1", "var1", "mean2", "var2"]
        # Try different numbers of walkers
        for N in [8, 10, 20]:
            sampler = EnsembleSampler(
                nwalkers=N,
                ndim=len(names),
                log_prob_fn=self.lnpdf_mixture,
                parameter_names=names,
            )
            coords = np.random.rand(N, len(names))
            lnps, _ = sampler.compute_log_prob(coords)
            assert len(lnps) == N
            assert lnps.dtype == np.float64

    def test_compute_log_prob_mixture_grouped(self):
        names = {"means": [0, 1], "vars": [2, 3], "constant": 4}
        # Try different numbers of walkers
        for N in [8, 10, 20]:
            sampler = EnsembleSampler(
                nwalkers=N,
                ndim=5,
                log_prob_fn=self.lnpdf_mixture_grouped,
                parameter_names=names,
            )
            coords = np.random.rand(N, 5)
            lnps, _ = sampler.compute_log_prob(coords)
            assert len(lnps) == N
            assert lnps.dtype == np.float64

    def test_run_mcmc(self):
        # Sort of an integration test
        n_walkers = 4
        sampler = EnsembleSampler(
            nwalkers=n_walkers,
            ndim=len(self.names),
            log_prob_fn=self.lnpdf,
            parameter_names=self.names,
        )
        guess = np.random.rand(n_walkers, len(self.names))
        n_steps = 50
        results = sampler.run_mcmc(guess, n_steps)
        assert results.coords.shape == (n_walkers, len(self.names))
        chain = sampler.chain
        assert chain.shape == (n_walkers, n_steps, len(self.names))



# -*- coding: utf-8 -*-

import warnings

import numpy as np
import pytest

from emcee import EnsembleSampler, backends

__all__ = ["test_blob_shape"]


class BlobLogProb(object):
    def __init__(self, blob_function):
        self.blob_function = blob_function

    def __call__(self, params):
        return -0.5 * np.sum(params ** 2), self.blob_function(params)


@pytest.mark.parametrize("backend", backends.get_test_backends())
@pytest.mark.parametrize(
    "blob_spec",
    [
        (True, False, 5, lambda x: np.random.randn(5)),
        (True, False, (5, 3), lambda x: np.random.randn(5, 3)),
        (True, False, (5, 3), lambda x: np.random.randn(1, 5, 1, 3, 1)),
        (True, False, 0, lambda x: np.random.randn()),
        (False, True, 2, lambda x: (1.0, np.random.randn(3))),
        (False, False, 0, lambda x: "face"),
        (False, False, 0, lambda x: object()),
        (False, False, 2, lambda x: ("face", "surface")),
        (False, True, 2, lambda x: (np.random.randn(5), "face")),
    ],
)
def test_blob_shape(backend, blob_spec):
    # HDF backends don't support the object type
    hdf_able, ragged, blob_shape, func = blob_spec
    if backend in (backends.TempHDFBackend,) and not hdf_able:
        return

    with backend() as be:
        np.random.seed(42)

        model = BlobLogProb(func)
        coords = np.random.randn(32, 3)
        nwalkers, ndim = coords.shape

        sampler = EnsembleSampler(nwalkers, ndim, model, backend=be)
        nsteps = 10

        if ragged:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", DeprecationWarning)
                sampler.run_mcmc(coords, nsteps)
        else:
            sampler.run_mcmc(coords, nsteps)

        shape = [nsteps, nwalkers]
        if isinstance(blob_shape, tuple):
            shape += blob_shape
        elif blob_shape > 0:
            shape += [blob_shape]

        assert sampler.get_blobs().shape == tuple(shape)
        if not hdf_able:
            assert sampler.get_blobs().dtype == np.dtype("object")


class VariableLogProb:
    def __init__(self):
        self.i = 3

    def __call__(self, *args):
        return 0, np.zeros(self.i)


@pytest.mark.parametrize("backend", backends.get_test_backends())
def test_blob_mismatch(backend):
    with backend() as be:
        np.random.seed(42)

        model = VariableLogProb()
        coords = np.random.randn(32, 3)
        nwalkers, ndim = coords.shape

        sampler = EnsembleSampler(nwalkers, ndim, model, backend=be)

        model.i += 1
        # We don't save blobs from the initial points
        # so blob shapes are taken from the first round of moves
        sampler.run_mcmc(coords, 1)

        model.i += 1
        with pytest.raises(ValueError):
            sampler.run_mcmc(coords, 1)



import os
import sys

import matplotlib.pyplot as pl
import numpy as np

# sys.path.prepend(os.path.abspath(os.path.join(__file__, "..", "..", "..")))
# import emcee


def plot_acor(acorfn):
    pass



# -*- coding: utf-8 -*-

try:
    import scipy
except ImportError:
    scipy = None
import pytest

from emcee import moves

from .test_proposal import _test_normal, _test_uniform

__all__ = ["test_normal_kde", "test_uniform_kde", "test_nsplits_kde"]


@pytest.mark.skipif(scipy is None, reason="scipy is not available")
def test_normal_kde(**kwargs):
    _test_normal(moves.KDEMove(), **kwargs)


@pytest.mark.skipif(scipy is None, reason="scipy is not available")
def test_uniform_kde(**kwargs):
    _test_uniform(moves.KDEMove(), **kwargs)


@pytest.mark.skipif(scipy is None, reason="scipy is not available")
def test_nsplits_kde(**kwargs):
    _test_normal(moves.KDEMove(nsplits=5), **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re
import sys

if len(sys.argv) <= 1:
    sys.exit(0)


def subber(m):
    return m.group(0).replace("``", "`")


prog = re.compile(r":(.+):``(.+)``")

for fn in sys.argv[1:]:
    print("Fixing links in {0}".format(fn))
    with open(fn, "r") as f:
        txt = f.read()
    txt = prog.sub(subber, txt)
    with open(fn, "w") as f:
        f.write(txt)
